#pragma once
#include <easy2d/e2dcommon.h>
#include <easy2d/e2dbase.h>

namespace easy2d
{


class Node;
class Loop;
class Sequence;
class Spawn;
class ActionManager;
class Sprite;


// 
class Action :
	public Object
{
	friend class ActionManager;
	friend class Loop;
	friend class Sequence;
	friend class Spawn;

public:
	Action();

	virtual ~Action();

	// ȡ״̬
	bool isRunning();

	// 
	void resume();

	// ͣ
	void pause();

	// ֹͣ
	void stop();

	// ȡ
	String getName() const;

	// ö
	void setName(
		const String& name
	);

	// ȡĿ
	virtual Action * clone() const = 0;

	// ȡĵת
	virtual Action * reverse() const = 0;

	// ö
	virtual void reset();

	// ȡöִĿ
	Node * getTarget() const;

	// ִжʱĿڵӸڵƳ
	void removeTargetWhenDone();

protected:
	// ʼ
	virtual void _init();

	// ¶
	virtual void _update();

	// ȡ״̬
	bool _isDone();

	// öʱ
	virtual void _resetTime();

	// ʼ
	void _startWithTarget(
		Node* target
	);

protected:
	String	_name;
	bool	_running;
	bool	_done;
	bool	_initialized;
	bool	_removeTarget;
	Node *	_target;
	float	_last;
};


// 
class FiniteTimeAction :
	public Action
{
public:
	// ضʱĳ
	explicit FiniteTimeAction(
		float duration
	);

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	float _duration;
	float _delta;
};


// λƶ
class MoveBy :
	public FiniteTimeAction
{
public:
	explicit MoveBy(
		float duration,	/* ʱ */
		Vector2 vector		/* ƶ */
	);

	// ȡöĿ
	virtual MoveBy * clone() const override;

	// ȡöĵת
	virtual MoveBy * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	Point	_startPos;
	Point	_prevPos;
	Vector2	_deltaPos;
};


// λƶ
class MoveTo :
	public MoveBy
{
public:
	explicit MoveTo(
		float duration,	/* ʱ */
		Point pos			/* Ŀ */
	);

	// ȡöĿ
	virtual MoveTo * clone() const override;

	// ȡöĵת
	virtual MoveTo * reverse() const override
	{
		E2D_WARNING(L"reverse() not supported in MoveTo");
		return nullptr;
	}

protected:
	// ʼ
	virtual void _init() override;

protected:
	Point _endPos;
};


// Ծ
class JumpBy :
	public FiniteTimeAction
{
public:
	explicit JumpBy(
		float duration,		/* ʱ */
		const Vector2& vec,		/* Ծ */
		float height,			/* Ծ߶ */
		int jumps = 1			/* Ծ */
	);

	// ȡöĿ
	virtual JumpBy * clone() const override;

	// ȡöĵת
	virtual JumpBy * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	Point	_startPos;
	Vector2	_deltaPos;
	float	_height;
	int		_jumps;
	Point	_prevPos;
};


// Ծ
class JumpTo :
	public JumpBy
{
public:
	explicit JumpTo(
		float duration,		/* ʱ */
		const Point& pos,		/* Ŀ */
		float height,			/* Ծ߶ */
		int jumps = 1			/* Ծ */
	);

	// ȡöĿ
	virtual JumpTo * clone() const override;

	// ȡöĵת
	virtual JumpTo * reverse() const override
	{
		E2D_WARNING(L"reverse() not supported in JumpTo");
		return nullptr;
	}

protected:
	// ʼ
	virtual void _init() override;

protected:
	Point _endPos;
};


// Ŷ
class ScaleBy :
	public FiniteTimeAction
{
public:
	explicit ScaleBy(
		float duration,	/* ʱ */
		float scale		/* Ա仯ֵ */
	);

	explicit ScaleBy(
		float duration,	/* ʱ */
		float scaleX,		/* Ա仯ֵ */
		float scaleY		/* Ա仯ֵ */
	);

	// ȡöĿ
	virtual ScaleBy * clone() const override;

	// ȡöĵת
	virtual ScaleBy * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	float	_startScaleX;
	float	_startScaleY;
	float	_deltaX;
	float	_deltaY;
};


// Ŷ
class ScaleTo :
	public ScaleBy
{
public:
	explicit ScaleTo(
		float duration,	/* ʱ */
		float scale		/* Ŀֵ */
	);

	explicit ScaleTo(
		float duration,	/* ʱ */
		float scaleX,		/* Ŀֵ */
		float scaleY		/* Ŀֵ */
	);

	// ȡöĿ
	virtual ScaleTo * clone() const override;

	// ȡöĵת
	virtual ScaleTo * reverse() const override
	{
		E2D_WARNING(L"reverse() not supported in ScaleTo");
		return nullptr;
	}

protected:
	// ʼ
	virtual void _init() override;

protected:
	float	_endScaleX;
	float	_endScaleY;
};


// ͸Խ䶯
class OpacityBy :
	public FiniteTimeAction
{
public:
	explicit OpacityBy(
		float duration,	/* ʱ */
		float opacity		/* Ա仯ֵ */
	);

	// ȡöĿ
	virtual OpacityBy * clone() const override;

	// ȡöĵת
	virtual OpacityBy * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	float _startVal;
	float _deltaVal;
};


// ͸Ƚ䶯
class OpacityTo :
	public OpacityBy
{
public:
	explicit OpacityTo(
		float duration,	/* ʱ */
		float opacity		/* Ŀֵ */
	);

	// ȡöĿ
	virtual OpacityTo * clone() const override;

	// ȡöĵת
	virtual OpacityTo * reverse() const override
	{
		E2D_WARNING(L"reverse() not supported in OpacityTo");
		return nullptr;
	}

protected:
	// ʼ
	virtual void _init() override;

protected:
	float _endVal;
};


// 붯
class FadeIn :
	public OpacityTo
{
public:
	// 붯
	explicit FadeIn(
		float duration		/* ʱ */
	)
	: OpacityTo(duration, 1) 
	{
	}
};


// 
class FadeOut :
	public OpacityTo
{
public:
	// 
	explicit FadeOut(
		float duration		/* ʱ */
	)
	: OpacityTo(duration, 0) 
	{
	}
};


// ת
class RotateBy :
	public FiniteTimeAction
{
public:
	explicit RotateBy(
		float duration,	/* ʱ */
		float rotation		/* Ա仯ֵ */
	);

	// ȡöĿ
	virtual RotateBy * clone() const override;

	// ȡöĵת
	virtual RotateBy * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	float _startVal;
	float _deltaVal;
};


// ת
class RotateTo :
	public RotateBy
{
public:
	explicit RotateTo(
		float duration,	/* ʱ */
		float rotation		/* Ŀֵ */
	);

	// ȡöĿ
	virtual RotateTo * clone() const override;

	// ȡöĵת
	virtual RotateTo * reverse() const override
	{
		E2D_WARNING(L"reverse() not supported in RotateTo");
		return nullptr;
	}

protected:
	// ʼ
	virtual void _init() override;

protected:
	float _endVal;
};


// ʱ
class Delay :
	public Action
{
public:
	explicit Delay(
		float duration	/* ӳʱ룩 */
	);

	// ȡöĿ
	virtual Delay * clone() const override;

	// ȡöĵת
	virtual Delay * reverse() const override;

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	float _delay;
	float _delta;
};


// ѭ
class Loop :
	public Action
{
public:
	explicit Loop(
		Action * action,	/* ִѭĶ */
		int times = -1		/* ѭ */
	);

	virtual ~Loop();

	// ȡöĿ
	virtual Loop * clone() const override;

	// ȡöĵת
	virtual Loop * reverse() const override;

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	Action * _action;
	int _times;
	int _totalTimes;
};


// ص
class CallFunc :
	public Action
{
public:
	explicit CallFunc(
		const Function<void()>& func /*  */
	);

	// ȡöĿ
	virtual CallFunc * clone() const override;

	// ȡöĵת
	virtual CallFunc * reverse() const override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

protected:
	Function<void()> _func;
};


// ˳
class Sequence :
	public Action
{
public:
	Sequence();

	explicit Sequence(
		const std::vector<Action*>& actions	/* б */
	);

	virtual ~Sequence();

	// ڽβӶ
	void add(
		Action * action
	);

	// ڽβӶ
	void add(
		const std::vector<Action*>& actions	/* б */
	);

	// ȡöĿ
	virtual Sequence * clone() const override;

	// ȡöĵת
	virtual Sequence * reverse() const;

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	UINT _currIndex;
	std::vector<Action*> _actions;
};


// ͬ
class Spawn :
	public Action
{
public:
	Spawn();

	explicit Spawn(
		const std::vector<Action*>& actions	/* б */
	);

	virtual ~Spawn();

	// ڽβӶ
	void add(
		Action * action
	);

	// ڽβӶ
	void add(
		const std::vector<Action*>& actions	/* б */
	);

	// ȡöĿ
	virtual Spawn * clone() const override;

	// ȡöĵת
	virtual Spawn * reverse() const;

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	std::vector<Action*> _actions;
};


// ؼ֡
class KeyFrame :
	public Object
{
public:
	KeyFrame(Image* image);

	KeyFrame(Image* image, const Rect& cropRect);

	virtual ~KeyFrame();

	// ȡ֡ͼƬ
	Image* getImage() const;

	// ֡ͼƬ
	void setImage(Image* image);

	// ȡü
	Rect getCropRect() const;

	// òü
	void setCropRect(const Rect& cropRect);

	// ȡ
	float getWidth() const;

	// ȡ߶
	float getHeight() const;

	// ȡС
	Size getSize() const;

private:
	Image*	_image;
	Rect	_cropRect;
};


// ֡
class FrameSequence :
	public Object
{
public:
	FrameSequence();

	explicit FrameSequence(
		const std::vector<KeyFrame*>& frames	/* ؼ֡ */
	);

	explicit FrameSequence(
		float interval							/* ֡룩 */
	);

	explicit FrameSequence(
		float interval,							/* ֡룩 */
		const std::vector<KeyFrame*>& frames	/* ؼ֡ */
	);

	virtual ~FrameSequence();

	// ӹؼ֡
	void add(
		KeyFrame* frame							/* ؼ֡ */
	);

	// ӹؼ֡
	void add(
		Image* image							/* ؼ֡ */
	);

	// Ӷؼ֡
	void add(
		const std::vector<KeyFrame*>& frames	/* ؼ֡б */
	);

	// Ӷؼ֡
	void add(
		const std::vector<Image*>& frames		/* ؼ֡б */
	);

	// ȡؼ֡
	const std::vector<KeyFrame*>& getFrames() const;

	// ȡؼ֡
	std::vector<KeyFrame*>& getFrames();

	// ȡ֡
	float getInterval() const;

	// ÿһ֡ʱ
	void setInterval(
		float interval		/* ֡룩 */
	);

	// ȡ֡Ŀ
	FrameSequence * clone() const;

	// ȡ֡ĵת
	FrameSequence * reverse() const;

protected:
	float _interval;
	std::vector<KeyFrame*> _frames;
};


// ֡
class Animation :
	public Action
{
public:
	Animation();

	explicit Animation(
		FrameSequence * seq
	);

	virtual ~Animation();

	// ȡ
	virtual FrameSequence * getFrameSequence() const;

	// ö
	virtual void setFrameSequence(
		FrameSequence * seq
	);

	// ȡöĿ
	virtual Animation * clone() const override;

	// ȡöĵת
	virtual Animation * reverse() const override;

	// ö
	virtual void reset() override;

protected:
	// ʼ
	virtual void _init() override;

	// ¶
	virtual void _update() override;

	// öʱ
	virtual void _resetTime() override;

protected:
	UINT _frameIndex;
	FrameSequence* _seq;
};


}