#pragma once
#include <easy2d/e2dnode.h>

namespace easy2d
{

class ShapeMaker;
class ShapeNode;

// ״
class Shape :
	public Object
{
	friend ShapeMaker;
	friend ShapeNode;

public:
	// ֱ
	static Shape* createLine(Point begin, Point end);

	// 
	static Shape* createRect(const Rect& rect);

	// ԲǾ
	static Shape* createRoundedRect(
		const Rect& rect,		// 
		const Vector2& radius	// Բǰ뾶
	);

	// Բ
	static Shape* createCircle(
		const Point& center,	// ԭ
		float radius			// 뾶
	);

	// Բ
	static Shape* createEllipse(
		const Point& center,	// ԭ
		const Vector2& radius	// 뾶
	);

	// 
	static Shape* createPolygon(
		std::initializer_list<Point> vertices	// ζ
	);

	// 
	static Shape* createPolygon(
		const Point* vertices,
		int count
	);

	// ȡаΧ
	Rect getBoundingBox(
		const Matrix32* transform = nullptr	// Ӧõ״ϵĶά任
	) const;

	// жͼǷ
	bool containsPoint(
		const Point& point,					// 
		const Matrix32* transform = nullptr	// ӦõϵĶά任
	) const;

	// ȡͼչһֱߵĳ
	float getLength() const;

	// ͼ
	float computeArea() const;

	// ͼϵλú
	bool computePointAtLength(
		float length,		// ͼϵλãΧ [0.0 - 1.0]
		Point& point,		// λ
		Vector2& tangent	// 
	) const;

protected:
	Shape(ID2D1Geometry* geo = nullptr);

	virtual ~Shape();

	void resetGeometry(ID2D1Geometry* geo);

protected:
	ID2D1Geometry* _geo;
};

// ״
class ShapeMaker : public Object
{
public:
	// ״ϲʽ
	enum class CombineMode
	{
		Union,      //  (A + B)
		Intersect,  //  (A + B)
		Xor,        // ԳƲ ((A - B) + (B - A))
		Exclude     //  (A - B)
	};

	// ϲ״
	static Shape* combine(
		Shape* shape1,
		Shape* shape2,
		CombineMode mode,					// ϲʽ
		const Matrix32* matrix = nullptr	// Ӧõ״
	);

	ShapeMaker() = default;

	virtual ~ShapeMaker();

	// ȡɵ״
	Shape* getShape() const;

	// ʼ·
	void beginPath(
		const Point& point = Point()	// ·ʼ
	);

	// ·
	void endPath(
		bool closed = false				// ·Ƿպ
	);

	// һ߶
	void addLine(const Point& point);

	// Ӷ߶
	void addLines(std::initializer_list<Point> points);

	// Ӷ߶
	void addLines(const Point* points, int count);

	// һη
	void addBezier(
		const Point& point1,	// ߵĵһƵ
		const Point& point2,	// ߵĵڶƵ
		const Point& point3		// ߵյ
	);

	// ӻ
	void addArc(
		const Point& point,		// յ
		const Size& radius,		// Բ뾶
		float rotation,			// ԲתǶ
		bool clockwise = true,	// Ƿ˳ʱ
		bool smallSize = true	// ǷȡС 180 Ļ
	);

protected:
	void openSink();

	void closeSink();

protected:
	Shape* _shape = nullptr;
	ID2D1GeometrySink* _sink = nullptr;
};


// ״
class ShapeNode :
	public Node
{
public:
	// ״ʽ
	enum class Style
	{
		Solid,		/*  */
		Round,		/*  */
		Fill,		/*  +  */
	};

public:
	// ֱ
	static ShapeNode* createLine(Point begin, Point end);

	// 
	static ShapeNode* createRect(
		const Size& size		// ο
	);

	// ԲǾ
	static ShapeNode* createRoundedRect(
		const Size& size,		// ο
		const Vector2& radius	// Բǰ뾶
	);

	// Բ
	static ShapeNode* createCircle(
		float radius			// 뾶
	);

	// Բ
	static ShapeNode* createEllipse(
		const Vector2& radius	// 뾶
	);

	// 
	static ShapeNode* createPolygon(
		std::initializer_list<Point> vertices	// ζ
	);

	// 
	static ShapeNode* createPolygon(
		const Point* vertices,
		int count
	);

	ShapeNode(Shape* shape = nullptr);

	virtual ~ShapeNode();

	// ȡ״
	Shape* getShape() const;

	// ״
	void setShape(Shape* shape);

	// ȡʽ
	Style getStyle() const;

	// ȡɫ
	Color getFillColor() const;

	// ȡɫ
	Color getStrokeColor() const;

	// ȡ
	float getStrokeWidth() const;

	// ɫ
	void setFillColor(
		Color fillColor
	);

	// ɫ
	void setStrokeColor(
		Color strokeColor
	);

	// 
	void setStrokeWidth(
		float strokeWidth
	);

	// ʽ
	void setStyle(Style style);

	// ཻʽ
	void setLineJoin(
		LineJoin lineJoin
	);

	virtual Rect getBounds() const override;

	virtual Rect getBoundingBox() const override;

	virtual bool containsPoint(Point const& point) const override;

	virtual void onRender() override;

protected:
	Style	_style;
	float	_strokeWidth;
	Color	_strokeColor;
	Color	_fillColor;
	Rect	_bounds;
	Shape* _shape;
	ID2D1StrokeStyle* _strokeStyle;
};

}